<?xml version="1.0"?>

<st-source>
<!-- Changes in change set PATCH-RA502912-FutureRecomp -->
<time-stamp>From VisualWorks®, 7.4.1 of May 30, 2006 on July 14, 2008 at 10:18:04 pm</time-stamp>


<class>
<name>PvSystemSupport</name>
<environment>PROVISO</environment>
<super>External.ExternalInterface</super>
<private>false</private>
<indexed-type>none</indexed-type>
<inst-vars></inst-vars>
<class-inst-vars></class-inst-vars>
<imports>
			PROVISO.PvSystemSupportDictionary.*
			</imports>
<category>PV-Support</category>
<attributes>
<includeFiles></includeFiles>
<includeDirectories></includeDirectories>
<libraryFiles></libraryFiles>
<libraryDirectories><item>d:\Proviso\development\Kila</item> <item>[unix]/usr/lib</item> <item>[win]$(windir)\system32</item></libraryDirectories>
<beVirtual type="eval">false</beVirtual>
<optimizationLevel type="sym">debug</optimizationLevel>
<package>PV-Base Extensions</package>
</attributes>
</class>

<class>
<name>PvWindowsSystemSupport</name>
<environment>PROVISO</environment>
<super>PROVISO.PvSystemSupport</super>
<private>false</private>
<indexed-type>none</indexed-type>
<inst-vars></inst-vars>
<class-inst-vars></class-inst-vars>
<imports>
			PROVISO.PvWindowsSystemSupportDictionary.*
			</imports>
<category>PV-Support</category>
<attributes>
<includeFiles><item>windef.h</item> <item>winnt.h</item> <item>winbase.h</item> <item>process.h</item></includeFiles>
<includeDirectories><item>d:\program files\microsoft visual studio\vc98\include</item></includeDirectories>
<libraryFiles><item>kernel32.dll</item></libraryFiles>
<libraryDirectories><item>d:\Proviso\development\Kila</item> <item>[unix]/usr/lib</item> <item>[win]$(windir)\system32</item></libraryDirectories>
<beVirtual type="eval">false</beVirtual>
<optimizationLevel type="sym">debug</optimizationLevel>
<package>PV-Base Extensions</package>
</attributes>
</class>

<methods>
<class-id>PROVISO.RecompServer</class-id> <category>transactions</category>

<body package="PV-DC-Recomputations" selector="computeScopeFor:on:">computeScopeFor: aTimeSeries on: recomputationActions
	"Determine the scope of the receiver's operation, which involves finding the period of the formula the given time series represents, and then finding the boundaries of that period."

	| startScope endScope scopeStartTimestamp scopeEndTimestamp |
	"Note - We have to do these in reverse order, because part of doing the scope calculation is to resync the formula itself into the model,
	and we need to end up with the original one in the model at the start time, not what was there at the end time..."
	endScope := recompEngine queryInterface
		getScopeFor: aTimeSeries
		timestamp: recomputationActions last timestamp.
	endScope isNil
		ifTrue: [
			self system 
				logDebugTwo: #RECOMP_ERROR
				text: 'No inputs for time series ', aTimeSeries printString, ' at or before ', (PvTimestamp fromSeconds: recomputationActions last timestamp) printString.
			^nil].

	startScope := recomputationActions size = 1
		ifTrue: [endScope]
		ifFalse: [
			self pvModel calendarManager resetLocalizedSchedules.
			recompEngine queryInterface
				getScopeFor: aTimeSeries
				timestamp: recomputationActions first timestamp].
	scopeStartTimestamp := startScope first key.
	scopeEndTimestamp := endScope first value.

	"Since we sometimes end up messing up localized schedules when we compute the scope, we need to reset them."
	self pvModel calendarManager resetLocalizedSchedules.

	^scopeStartTimestamp -&gt; scopeEndTimestamp</body>

<body package="PV-DC-Recomputations" selector="handleTupleContainer:">handleTupleContainer: aTupleContainer 

	| theLabel |
	"Reject dups sent from TSE but allow dups sent by me"
	(self isDuplicateContainer: aTupleContainer) 
		ifTrue: [^self rejectDuplicateContainer: aTupleContainer].

	aTupleContainer isForRecomputations
		ifFalse: [
			(RecompCommandLog on: self) transactionDo: [:txn | 
				self releaseNonRecompTupleContainer: aTupleContainer in: txn.
				self updateLastContainerTimestamp: aTupleContainer timestamp in: txn].
			^nil].

	"Now we may need to wait for the CMEs to catch up..."
	[self isReadyForTupleContainer: aTupleContainer]
		whileFalse: [(Delay forSeconds: 60) wait].

	self system extendedStatus: 'Tuple Processing'.
	"Process the entire tuple container as an xlog"
	self logAboutToProcesesTupleContainer: aTupleContainer.
	theLabel := aTupleContainer label.
	(RecompCommandLog on: self) transactionDo: 
		[:txn | | resultsFn bufferFn sourceFn tempFn |
		self releaseProcessedTupleContainer: aTupleContainer in: txn.
		resultsFn := self fileInterface recompResultsStateFilename.
		sourceFn := self fileInterface recompSourceStateFilename.
		tempFn := self fileInterface recompTempFilename.
		bufferFn := self fileInterface outputBufferForTseStateFilename.
		(self isMyContainer: aTupleContainer)
			ifTrue: [
				| tuples |
				tuples := OrderedCollection new.
				aTupleContainer do: [:each | tuples add: each].
				self fileInterface saveRecompTuples: tuples to: tempFn. 
				txn recomputeAll: tempFn saveResultsTo: resultsFn saveActionsTo: sourceFn originLabel: theLabel.
				self handlePostRecomputeIn: txn for: resultsFn and: sourceFn to: bufferFn on: theLabel.
				txn deleteBufferedRecompResult: tempFn]
			ifFalse: [
				aTupleContainer do: [:tuple | 
					tuple isHeartbeat
						ifFalse: [
							txn recompute: tuple saveResultsTo: resultsFn saveActionsTo: sourceFn originLabel: theLabel.
							self handlePostRecomputeIn: txn for: resultsFn and: sourceFn to: bufferFn on: theLabel]]].
		txn outputBufferedRecompResultsToTse: bufferFn.
		txn deleteBufferedRecompResult: bufferFn.
		self updateLastContainerTimestamp: aTupleContainer timestamp in: txn].

	self system extendedStatus: 'Idle'.</body>

<body package="PV-DC-Recomputations" selector="isReadyForTupleContainer:">isReadyForTupleContainer: aTupleContainer
	"Figure out if we can do the recomputation action(s) in the given tuple container. 
	Return true or false to answer the question."

	| recomputationActions affectedTimeSeries cmeTimes referenceDateString returnValue farthestDate |
	"If we don't have any recomp actions, we're done. This should never happen, since we've already tested for this."
	recomputationActions := (aTupleContainer allTuples
		select: [:each | each isRecomputationAction]) asOrderedCollection.

	recomputationActions isEmpty
		ifTrue: [^true].

	"For each affected time series, find the labels, and then ask each one where it is at in the timeline."
	cmeTimes := Dictionary new.
	affectedTimeSeries := recompEngine queryInterface getAffectedTimeSeriesFor: recomputationActions first.
	affectedTimeSeries isEmpty
		ifTrue: [^true].
	farthestDate := Infinity negative.
	affectedTimeSeries do: [:eachTimeSeries |
		| labels scope |
		scope := self computeScopeFor: eachTimeSeries on: recomputationActions.
		scope value: (scope value min: recomputationActions first timestamp).
		farthestDate := farthestDate max: scope value.
		labels := PvTimeSeriesBuildCommand
			labelsForMetricId: eachTimeSeries metricId
			subelementId: eachTimeSeries resourceId
			between: scope key
			and: scope value.
		labels do: [:eachLabel |
			| cmePair |
			cmePair := CMETimeSeries new getApplicationForOperatorLabel: eachLabel asNumber.
			(cmePair value notNil and: [(cmeTimes includesKey: cmePair key) not])
				ifTrue: [
					| cmeApplication currentTime |
					cmeApplication := cmePair value.
					[currentTime := cmeApplication server streamLabelOperator periodPolicy last]
						on: Error do: [:signal | signal return: nil].
					cmeTimes at: cmePair key put: currentTime]]].

	"Compare each label's current position with the scope, and log any that aren't far enough along."
	referenceDateString := (PvTimestamp fromSeconds: farthestDate) printString.
	returnValue := true.
	cmeTimes keysAndValuesDo: [:eachServerName :eachTime |
		eachTime isNil 
			ifTrue: [
				self system 
					logDebugTwo: #RECOMP_NOT_READY
					text: eachServerName, ' is at at an unknown time and needs to be past ', referenceDateString, ' in order to continue with this recomputation.'.
				returnValue := false]
			ifFalse: [
				eachTime &lt; farthestDate
					ifTrue: [
						self system 
							logDebugTwo: #RECOMP_NOT_READY
							text: eachServerName, ' is at ', (PvTimestamp fromSeconds: eachTime) printString, ' and needs to be past ', referenceDateString, ' in order to continue with this recomputation.'.
						returnValue := false]]].

	^returnValue</body>
</methods>


<methods>
<class-id>PROVISO.RecomputationEngine</class-id> <category>building</category>

<body package="PV-DC-Recomputations" selector="buildTimeSeriesFromCommand:">buildTimeSeriesFromCommand: command
	"Note - command must be an instance of PvTimeSeriesBuildCommand."
	
	| id action formula subelement |
	formula := self pvModel formulaOrLocalizedScheduleWithId: command metricId.
	formula isNil ifTrue: [
		self logMissingFormulaFromCommand: command.
		^self].
	formula formulaPrototype ifNotNil: [:formulaPrototype |
		formulaPrototype
			deduplicate: false;
			deduplicateIntermediate: false].
	self queryInterface resyncSubelementWithId: command resourceId.
	subelement := self pvModel resourceWithId: command resourceId.
	"There is a case when a SE gets deleted, PBL doesn't remove the time series until the start of the next hour. If we try and do recomp in that time interval, we can end up with nil instead of a valid SE. In this case, we can just ignore the time series..."
	subelement isNil
		ifTrue: [^self].

	self queryInterface resyncPropertiesFor: formula resource: subelement atOrBefore: scopeStartTimestamp.

	id := (MetricSeriesIdentifier 
		metric: formula
		resource: (self pvModel subelementManager itemWithId: command resourceId))
			msid: command msid;
			yourself.
	action := command resyncAction.
	" This is to cover test case scenarios where the actions comes as a symbol "
	((action isSymbol and: [action = #D]) or: [action = 'D']) 
		ifTrue: [self deleteTimeSeriesForId: id]
		ifFalse: ["add the timeseries if it isn't already there"
			self 
				addOrUpdateTimeSeriesForId: id
				fromCommand: command]</body>
</methods>

<methods>
<class-id>PROVISO.RecomputationEngine</class-id> <category>logging</category>

<body package="PV-DC-Recomputations" selector="logMissingFormulaFromCommand:">logMissingFormulaFromCommand: command

	self system 
		logWarning: #NO_FORMULA
		messageId: #DC10437
		template: 'No formula for: &lt;1p&gt;'
		arguments: (Array with: command)</body>
</methods>

<methods>
<class-id>PROVISO.RecomputationEngine</class-id> <category>recomputation</category>

<body package="PV-DC-Recomputations" selector="computeScopeFor:">computeScopeFor: aTimeSeries
	"Determine the scope of the receiver's operation, which involves finding the period of the formula the given time series represents, and then finding the boundaries of that period."

	| startScope endScope timeBlock |
	"Note - We have to do these in reverse order, because part of doing the scope calculation is to resync the formula itself into the model,
	and we need to end up with the original one in the model at the start time, not what was there at the end time..."
	endScope := queryInterface
		getScopeFor: aTimeSeries
		timestamp: recomputationActions last timestamp.
	endScope isNil
		ifTrue: [
			self system 
				logDebugTwo: #RECOMP_ERROR
				text: 'No inputs for time series ', aTimeSeries printString, ' at or before ', (PvTimestamp fromSeconds: recomputationActions last timestamp) printString.
			^nil].

	startScope := recomputationActions size = 1
		ifTrue: [endScope]
		ifFalse: [
			self pvModel calendarManager resetLocalizedSchedules.
			queryInterface
				getScopeFor: aTimeSeries
				timestamp: recomputationActions first timestamp].
	scopeStartTimestamp := startScope first key.
	scopeEndTimestamp := endScope first value.
	periodStartTimestamp := startScope last key.
	periodEndTimestamp := endScope last value.

	timeBlock := [:timestamp |
		(timestamp == Infinity positive) | (timestamp == Infinity negative)
			ifTrue: ['']
			ifFalse: [' (', (PvTimestamp fromSeconds: timestamp) printString, ')']].

	self system 
		logDebugTwo: #RECOMP_SCOPE
		text: ('Scope: %1%2 -&gt; %3%4'
			bindWith: scopeStartTimestamp printString
			with: (timeBlock value: scopeStartTimestamp)
			with: scopeEndTimestamp printString
			with: (timeBlock value: scopeEndTimestamp)).

	self system 
		logDebugTwo: #RECOMP_PERIOD
		text: ('Scope: %1%2 -&gt; %3%4'
			bindWith: periodStartTimestamp printString
			with: (timeBlock value: periodStartTimestamp)
			with: periodEndTimestamp printString
			with: (timeBlock value: periodEndTimestamp)).

	"Since we sometimes end up messing up localized schedules when we compute the scope, we need to reset them."
	self pvModel calendarManager resetLocalizedSchedules.</body>

<body package="PV-DC-Recomputations" selector="gatherOriginalResultsFor:passing:">gatherOriginalResultsFor: aTimeSeries passing: recompActions
	"Retrieve the original metrics produced by the formula assocaited with the given time series, during the time periond in question."

	originalResults := queryInterface
		getMetricsForMid: aTimeSeries metricId
		rid: aTimeSeries resourceId
		between: scopeStartTimestamp
		and: scopeEndTimestamp
		passing: recompActions.</body>

<body package="PV-DC-Recomputations" selector="recomputeAll:originLabel:">recomputeAll: aCollection originLabel: originLabel
	"Run recomputation using the given recomputation action."

	| recomputationRecords actions |
	self system extendedStatus: 'Running Recomputation'.
	recomputationRecords := self removeNonRecomputationRecordsFrom: aCollection.
	actions := self removeDuplicateRecordsFrom: recomputationRecords using: originLabel.

	actions doWithIndex: [:each :index |
		self system
			logDebugTwo: #RECOMP
			text: 'Processing recomp record: ', each printString, ' - ', index printString, ' of ', actions size printString].

	recomputationActions := self convertRecomputationActions: actions.
	recomputationActions isEmpty
		ifTrue: [^#(() ())].
	recomputationActions := recomputationActions asSortedCollection: [:a :b | a timestamp &lt;= b timestamp].

	deltas := OrderedCollection new.
	self affectedTimeSeriesDo: [:eachTimeSeries |
		self computeScopeFor: eachTimeSeries.
		scopeStartTimestamp notNil
			ifTrue: [
				self
					computeInputsFor: eachTimeSeries;
					computeMetadataChangesFor: eachTimeSeries;
					computeNonInputMetadataChangesFor: eachTimeSeries;
					gatherOriginalResultsFor: eachTimeSeries passing: actions;
					gatherMetricsFor: eachTimeSeries;
					runReplayFor: eachTimeSeries]].

	self switchDeltasUsing: recomputationRecords.
	self system extendedStatus: 'Done Recomputation'.
	^Array with: deltas with: recomputationRecords</body>
</methods>


<methods>
<class-id>PROVISO.SharedPriorityQueue</class-id> <category>access operations</category>

<body package="PV-Base Extensions" selector="contents">contents
	"&lt;OrderedCollection&gt; Answer the contents of the queue" 

	^[accessProtect critical: [contents copy]] valueUninterruptably</body>
</methods>


<methods>
<class-id>PROVISO.TupleStorage</class-id> <category>tuple queries - for recomp</category>

<body package="PV-Tuple Storage Engine" selector="getMetricsForMid:rid:between:and:adjTime:">getMetricsForMid: metricId rid: resourceId between: startTsNumber and: endTsNumber adjTime: adjTime
	"&lt;Array&gt; Ask tuple storage for the set of metric records in the given timeframe, with the given mid/rid."
	
	^self getMetricsForMid: metricId rid: resourceId between: startTsNumber and: endTsNumber adjTime: adjTime passing: #()</body>

<body package="PV-Tuple Storage Engine" selector="getMetricsForMid:rid:between:and:adjTime:passing:">getMetricsForMid: metricId rid: resourceId between: startTsNumber and: endTsNumber adjTime: adjTime passing: recompActions
	"&lt;Array&gt; Ask tuple storage for the set of metric records in the given timeframe, with the given mid/rid."
	
	| strm |
	[strm := self
		getTuplesForMid: metricId
		rid: resourceId
		from: (self decodeInfinity: startTsNumber)
		to: (self decodeInfinity: endTsNumber)
		adjTime: (self decodeInfinity: adjTime)
		count: Infinity positive
		dbSession: self dbSession
		passing: recompActions.
	^strm upToEnd]
		ensure: [strm ifNotNil: [strm close]]</body>
</methods>

<methods>
<class-id>PROVISO.TupleStorage</class-id> <category>tuple queries - tuple storage</category>

<body package="PV-Tuple Storage Engine" selector="handleRemoteQueryForMid:rid:from:to:on:for:passing:">handleRemoteQueryForMid: mid rid: rid from: startTime to: endTime on: label for: agent passing: recompActions

	| cmePair remoteStream |
	cmePair := CMETimeSeries new getApplicationForOperatorLabel: label.
	cmePair value isNil
		ifFalse: [
			| cmeApplication |
			cmeApplication := cmePair value.
			[
				remoteStream := RemoteStreamProxy on: (cmeApplication server streamLabelOperator
					protectedHandleRecomputationRecords: recompActions
					whileGettingTuplesForMid: mid
					rid: rid 
					from: startTime 
					to: endTime).
				remoteStream finalizeAction: #destroy]
			on: Error 
			do: [ :ex | 
				self logRemoteQueryFailed: ex label: label.
				ex pass]].

	^remoteStream</body>

<body package="PV-Tuple Storage Engine" selector="tsGetTuplesForMid:rid:from:to:">tsGetTuplesForMid: mid rid: rid from: startTime to: endTime

	| labels |
	labels := self planAccessor labelsForMetricId: mid subelementId: rid between: startTime and: endTime.
	^self
		tsGetTuplesForMid: mid 
		rid: rid 
		from: startTime 
		to: endTime 
		on: labels 
		passing: #()</body>

<body package="PV-Tuple Storage Engine" selector="tsGetTuplesForMid:rid:from:to:on:passing:">tsGetTuplesForMid: mid rid: rid from: startTime to: endTime on: labels passing: recompActions

	| agents results |
	(mid isNil and: [recompActions isEmpty])
		ifTrue: [^self].
	agents := labels collect: [ :label | label -&gt; (self lookupNode: label ifAbsent: [nil])].
	agents := agents select: [ :ea | ea value notNil].
	results := OrderedCollection new.
	agents do:
		[ :assn |
		| agent label |
		agent := assn value.
		label := assn key.
		agent isRemote
			ifTrue: [
				| value |
				self logRemoteQuery: assn key agent: agent mid: mid rid: rid from: startTime to: endTime.
				value := self
					handleRemoteQueryForMid: mid
					rid: rid 
					from: startTime 
					to: endTime 
					on: label 
					for: agent 
					passing: recompActions.
				value notNil ifTrue: [results add: value]]
			ifFalse: [results add: (agent getTuplesForMid: mid rid: rid from: startTime to: endTime)]].
	^mid isNil
		ifTrue: [self]
		ifFalse: [
			(self directionFrom: startTime to: endTime) &gt;= 0
				ifTrue: [MergeSortStream forTuplesOnAll: results]
				ifFalse: [MergeSortStream forReverseTuplesOnAll: results]]</body>
</methods>

<methods>
<class-id>PROVISO.TupleStorage</class-id> <category>tuple queries - unioned</category>

<body package="PV-Tuple Storage Engine" selector="getDbPositionForType:on:">getDbPositionForType: tupleType on: labels

	| stream timestamps |
	labels isEmpty
		ifTrue: [^nil].

	stream := (WriteStream on: (String new: 100))
		nextPutAll: 'select 
  max(s.dte_last_load_date) as startOfLastLoadedPeriod
from chnl_table_set_type t, chnl_table_load_status s 
where t.idx_ind = s.IDX_TABLE_SET_TYPE and t.str_name = ';
		nextPutAll: tupleType printString;
		nextPutAll: ' and s.idx_subchnl in (';
		yourself.

	labels
		do: [:each | stream print: each]
		separatedBy: [stream nextPutAll: ', '].

	stream nextPutAll:  ')'.

	timestamps := (self dbSession execute: stream contents) collect: [:each | each first].
	timestamps := timestamps reject: [:each | each isNil].
	^timestamps isEmpty
		ifTrue: [nil]
		ifFalse: [
			timestamps
				inject: timestamps first
				into: [:sum :each | sum max: each]].</body>

<body package="PV-Tuple Storage Engine" selector="getDbPositionOn:">getDbPositionOn: labels

	| sparseTimestamp periodicTimestamp |
	sparseTimestamp := self getDbPositionForType: 'SNRW' on: labels.
	periodicTimestamp := self getDbPositionForType: 'NRAW' on: labels.

	"We add 3599 to the timestamp, since the DB column we're reading is start of period..."
	sparseTimestamp isNil
		ifTrue: [^periodicTimestamp].
	periodicTimestamp isNil
		ifTrue: [^sparseTimestamp].
	^(sparseTimestamp min: periodicTimestamp) + 3599</body>

<body package="PV-Tuple Storage Engine" selector="getTuplesForMid:rid:from:to:adjTime:count:dbSession:">getTuplesForMid: mid rid: rid from: startTime to: stopTime adjTime: adjTime count: count dbSession: dbSession
	"&lt;ReadStream&gt; Get all the tuples from tuples storage and the database for the given mid/rid.  Order by tuple timestamp (startTime may be greater than stopTime to indicate reverse order).  The startTime and endTime can be infinite to formulate a query that gets all tuples. The count is used to optimize the database queries, but does not restrict the number of tuples that can be in the resulting read stream (do not read more than count tuples from the resulting read stream, the results may not be accurate)."

	^self
		getTuplesForMid: mid
		rid: rid 
		from: startTime 
		to: stopTime 
		adjTime: adjTime 
		count: count 
		dbSession: dbSession 
		passing: #()</body>

<body package="PV-Tuple Storage Engine" selector="getTuplesForMid:rid:from:to:adjTime:count:dbSession:passing:">getTuplesForMid: mid rid: rid from: startTime to: stopTime adjTime: adjTime count: count dbSession: dbSession passing: recompActions
	"&lt;ReadStream&gt; Get all the tuples from tuples storage and the database for the given mid/rid.  Order by tuple timestamp (startTime may be greater than stopTime to indicate reverse order).  The startTime and endTime can be infinite to formulate a query that gets all tuples. The count is used to optimize the database queries, but does not restrict the number of tuples that can be in the resulting read stream (do not read more than count tuples from the resulting read stream, the results may not be accurate)."

	| labels transitionTime tupleStreams start stop adjustmentStart adjustmentStop dbStart dbStop mergeSort direction needTuplesFromTS |
	start := startTime ifNil: [Infinity negative].
	stop := stopTime ifNil: [Infinity positive].
	direction := self directionFrom: startTime to: stopTime ifIndeterminate: [^#() readStream].

	"Determine the earliest time for which TSE has tuples for the given mid/rid"
	labels := self planAccessor labelsForMetricId: mid subelementId: rid between: startTime and: stopTime.
	transitionTime := self getDbPositionOn: labels.
	transitionTime isNil
		ifTrue: [transitionTime := start].

	"If the latest DB time is before the stop time, we need to talk to the CME(s) as well."
	needTuplesFromTS := transitionTime &lt; stop.

	"Now figure out the time ranges for the DB ajustment and tuple queries"
	needTuplesFromTS
		ifTrue: [
			start &lt;= stop
				ifTrue: [
					adjustmentStart := transitionTime + 1 max: start.
					adjustmentStop := stop.
					dbStart := start.
					dbStop := transitionTime]
				ifFalse: [
					adjustmentStart := start.
					adjustmentStop := transitionTime - 1 max: stop.
					dbStart := transitionTime.
					dbStop := stop]]
		ifFalse: [
			dbStart := start.
			dbStop := stop].

	tupleStreams := OrderedCollection new.

	"1) First, lets go after the adjusted tuples from the database...do this first because we want the adjustments to shadow other results (if there are no tuples in tuple storage, no need to do this)."
	needTuplesFromTS 
		ifTrue: [
			tupleStreams add: (self 
				dbGetAdjustedSparseTuplesForMid: mid 
				rid: rid 
				from: adjustmentStart 
				to: adjustmentStop 
				adjTime: adjTime 
				count: count 
				dbSession: dbSession).
			"For now, we don't need the following because we don't support adjustments to periodic metrics"
			"tupleStreams add:
				(self dbGetAdjustedTuplesForMid: mid rid: rid from: adjustmentStart to: adjustmentStop adjTime: adjTime count: count dbSession: dbSession."].

	"2) Next go after tuples outside the Tuple Storage window in the DB"
	tupleStreams add: (self 
		dbGetSparseTuplesForMid: mid 
		rid: rid 
		from: dbStart
		to: dbStop
		adjTime: adjTime 
		count: count 
		dbSession: dbSession).
	tupleStreams add: (self
		dbGetTuplesForMid: mid
		rid: rid 
		from: dbStart 
		to: dbStop 
		adjTime: adjTime 
		count: count 
		dbSession: dbSession).

	"3) Finally, get the tuples from tuple storage (if there aren't any tuples in tuple storage, skip this)"
	needTuplesFromTS 
		ifTrue: [
			tupleStreams add: (self 
				tsGetTuplesForMid: mid 
				rid: rid 
				from: adjustmentStart 
				to: adjustmentStop
				on: labels
				passing: recompActions)]
		ifFalse: [
			self 
				tsGetTuplesForMid: nil 
				rid: nil 
				from: nil 
				to: nil
				on: labels
				passing: recompActions].

	direction &gt;= 0
		ifTrue: [mergeSort := MergeSortStream forTuplesOnAll: tupleStreams]
		ifFalse: [mergeSort := MergeSortStream forReverseTuplesOnAll: tupleStreams].
	^DeduplicatingStream 
		on: mergeSort
		comparisonBlock: [ :a :b | a timestamp = b timestamp]</body>
</methods>


<methods>
<class-id>PROVISO.MockRecomputationQueryInterface</class-id> <category>queries</category>

<body package="PV-DC-Recomputations" selector="getMetricsForMid:rid:between:and:passing:">getMetricsForMid: metricId rid: resourceId between: startTsNumber and: endTsNumber passing: recompActions

	^metricsHistory select: [:each |
		each metricId = metricId
			and: [each resourceId = resourceId]
			and: [each timestamp between: startTsNumber and: endTsNumber]]</body>
</methods>


<methods>
<class-id>PROVISO.StreamLabelProcessor</class-id> <category>processing (recomputations)</category>

<body package="PV-CME-Core" selector="getNearRealTimeRecordsForMid:rid:from:to:">getNearRealTimeRecordsForMid: mid rid: rid from: startTime to: endTime

	^self getMetricRecordsBetween: startTime  and: endTime forMetricId: mid andResourceId: rid</body>

<body package="PV-CME-Core" selector="protectedHandleRecomputationRecords:whileGettingTuplesForMid:rid:from:to:">protectedHandleRecomputationRecords: recompActions whileGettingTuplesForMid: mid rid: rid from: startTime to: endTime

	| answer tsRecords nrtRecords |
	[
		| timeSeries tupleStorageNode needToUseTS |
		mid notNil
			ifTrue: [
				"First, see if we have any NRT records"
				timeSeries := self tupleRepository atMid: mid rid: mid ifAbsent: [nil].
				nrtRecords := (timeSeries notNil and: [timeSeries shouldStore])
					ifTrue: [self getNearRealTimeRecordsForMid: mid rid: rid from: startTime to: endTime]
					ifFalse: [nil].
				nrtRecords isNil
					ifTrue: [needToUseTS := true]
					ifFalse: [
						| earliestRecord |
						earliestRecord := nrtRecords first.
						"If there's a possible gap between NRT and the startTime, we have to go to TS..."
						needToUseTS := earliestRecord timestamp &gt; startTime].
				needToUseTS
					ifTrue: [
						tupleStorageNode := self tupleStorage lookupNode: self collectorNumber ifAbsent: [nil].
						tsRecords := tupleStorageNode isNil
							ifTrue: [nil]
							ifFalse: [tupleStorageNode remoteGetTuplesForMid: mid rid: rid from: startTime to: endTime]]].
		recompActions do: [:each | self handleRecomputationRecord: each]] valueUninterruptably.

	answer := OrderedCollection new.
	tsRecords notNil
		ifTrue: [answer addAll: tsRecords].
	nrtRecords notNil
		ifTrue: [answer addAll: nrtRecords].

	^ReadStream on: answer</body>
</methods>


<methods>
<class-id>PROVISO.RecomputationQueryInterface</class-id> <category>queries</category>

<body package="PV-DC-Recomputations" selector="getMetricsForMid:rid:between:and:">getMetricsForMid: metricId rid: resourceId between: startTsNumber and: endTsNumber 
	"Ask tuple storage for the set of metric records in the given timeframe, with the given mid/rid."
	
	^self getMetricsForMid: metricId rid: resourceId between: startTsNumber and: endTsNumber passing: #()</body>

<body package="PV-DC-Recomputations" selector="getMetricsForMid:rid:between:and:passing:">getMetricsForMid: metricId rid: resourceId between: startTsNumber and: endTsNumber passing: passingTuples
	"Ask tuple storage for the set of metric records in the given timeframe, with the given mid/rid."
	
	^self 
		queryTupleStorageWith: #getMetricsForMid:rid:between:and:adjTime:passing:
		arguments: 
			(Array 
				with: metricId
				with: resourceId
				with: (self encodeInfinity: startTsNumber)
				with: (self encodeInfinity: endTsNumber)
				with: '+infinity'
				with: passingTuples)</body>
</methods>


<methods>
<class-id>PROVISO.RecompTupleInjectionInterface</class-id> <category>initialize-release</category>

<body package="PV-DC-Recomputations" selector="initialize">initialize
	
	super initialize.
	self agentQueue: (SharedPriorityQueue sortBlock: [:container1 :container2 | 
		container1 isForRecomputations 
			ifTrue: [
				container2 isForRecomputations 
					ifTrue: [container1 timestamp &lt; container2 timestamp]
					ifFalse: [true]]
				ifFalse: [
					container2 isForRecomputations 
						ifTrue: [false]
						ifFalse: [container1 timestamp &lt; container2 timestamp]]])</body>
</methods>



</st-source>
