<?xml version="1.0"?>

<st-source>
<time-stamp>From VisualWorks®, 7.4.1 of May 30, 2006 on July 14, 2008 at 2:28:03 pm</time-stamp>
<!-- Package Changes for PV-DC-Plan-Builder(Kila.89,apikovsky)* -->


<methods>
<class-id>PROVISO.PlanBuilder</class-id> <category>plan building</category>

<body package="PV-DC-Plan-Builder" selector="handleBuiltGenericTimeSeries:for:">handleBuiltGenericTimeSeries: genericTs for: specificTs 
	"A ts for a generic formula should have the same label as the specific formula's ts"	

	| cNum labelToAssign specificLabels | 
	cNum := self getCollectorNumberForTimeSeries: genericTs.
	labelToAssign :=
		(cNum isNil or: [cNum = 0])
			ifTrue: [self operatorLabelZero]
			ifFalse: [self collectorLabelForId: cNum].

	specificLabels := Set new.
	genericTs formula specificFormulas do: 
		[:specificFormula | 
		(self atMid: specificFormula id rid: genericTs resourceId) ifNotNil: 
			[:ts | specificLabels add: ts streamLabel]].
	specificLabels size = 1
		ifTrue: [labelToAssign := specificLabels anyOne]
		ifFalse: 
			[specificLabels size &gt; 1 
				ifTrue: [self logMultipleSpecific: specificLabels toGeneric: genericTs]].

	self 
		changeLabelForTimeSeries: genericTs 
		to: labelToAssign</body>
</methods>

<methods>
<class-id>PROVISO.PlanBuilder</class-id> <category>logging</category>

<body package="PV-DC-Plan-Builder" selector="logMultipleSpecific:toGeneric:">logMultipleSpecific: specificLabels toGeneric: aTimeSeries

	| specificFormulas |
	specificFormulas := (aTimeSeries specificFormulas collect: [:e | e nameString]) asArray.
	self system
		logWarning: #MULTIPLE_SPECIFIC_TO_GENERIC
		messageId: #DC10010
		text: 'There are multiple specific formulas: ', specificFormulas printString ,' for this generic formula: ' , aTimeSeries id printString , '. These specific formulas are for the same resource and reside in different labels ', specificLabels asArray printString.</body>

<body package="PV-DC-Plan-Builder" selector="logSavePlanError">logSavePlanError
	&lt;comment: 'Stopping plan builder because plan save to db failed' 
	probableCause: 'Could be duplicate rows, database error, Smalltalk exception, or something else'
	suggestedAction: 'Restart the plan builder'&gt;

	self system 
		logWarning: #PLAN_SAVE_ERROR
		messageId: #DC10009
		text: 'Stopping plan builder because plan save to db failed'.</body>
</methods>

<methods>
<class-id>PROVISO.PblPlanFromDatabaseBuilder</class-id> <category>private</category>

<body package="PV-DC-Plan-Builder" selector="buildResourceBuckets">buildResourceBuckets
	
	|  tsCount inputCount tsWalkbacksWritten inputWalkbacksWritten |

	tsCount := 0.
	tsWalkbacksWritten := 0.
	PlanAccessor new 
		planVersionsPast: self lastLoadedPlanVersion
		timeSeriesDo: 
			[:eachRow | 
			[self buildTimeSeriesFromDbRow: eachRow.
			tsCount := tsCount + 1] 
				on: Error
				do: 	
					[:ex | 
					self logErrorBuildTsErrorFrom: eachRow reason: ex.
					tsWalkbacksWritten &gt; 2 
						ifFalse: 
							[PvWalkbackWriter new dumpWalkbackForException: ex.
							tsWalkbacksWritten := tsWalkbacksWritten + 1]]].
	self system logDebugTwo: #PLAN_STAT text: 'Loaded ' , tsCount asString , ' time series from db'.

"Load inputs"
	inputCount := 0.
	inputWalkbacksWritten := 0.
	PlanAccessor new 
		planVersionsPast: self lastLoadedPlanVersion
		inputsDo: 
			[:eachRow | 
			[self buildInputFromDbRow: eachRow.
			inputCount := inputCount + 1] 
				on: Error
				do: 
					[:ex | 
					self logErrorBuildInputErrorFrom: eachRow reason: ex.
					inputWalkbacksWritten &gt; 2 
						ifFalse: 
							[PvWalkbackWriter new dumpWalkbackForException: ex.
							inputWalkbacksWritten := inputWalkbacksWritten + 1]]].
	self system logDebugTwo: #PLAN_STAT text: 'Loaded ' , inputCount asString , ' inputs from db'.

	"Actually assign each ts it's label"
	self rawData do: 
		[:ts | 
		ts streamLabel assignToTimeSeries: ts]</body>
</methods>

<methods>
<class-id>PROVISO.PblPlanFromDiskBuilder</class-id> <category>As yet unclassified</category>

<body package="PV-DC-Plan-Builder" selector="buildResourceBuckets">buildResourceBuckets

	| tsCount inputCount tsWalkbacksWritten inputWalkbacksWritten |
	

"Load timeSeries"
	tsCount := 0.
	tsWalkbacksWritten := 0.
	self planBuilder planAndMetadataDumperLoader 
		timeSeriesDo: 
			[:eachRow | 
			[self buildTimeSeriesFromDbRow: eachRow.
			tsCount := tsCount + 1] 
				on: Error
				do: 	
					[:ex | 
					self logErrorBuildTsErrorFrom: eachRow reason: ex.
					tsWalkbacksWritten &gt; 2 
						ifFalse: 
							[PvWalkbackWriter new dumpWalkbackForException: ex.
							tsWalkbacksWritten := tsWalkbacksWritten + 1]]].
	self system logDebugTwo: #PLAN_STAT text: 'Loaded ' , tsCount asString , ' time series from disk'.

"Load inputs"
	inputCount := 0.
	inputWalkbacksWritten := 0.
	self planBuilder planAndMetadataDumperLoader 
		inputsDo: 
			[:eachRow | 
			[self buildInputFromDbRow: eachRow.
			inputCount := inputCount + 1] 
				on: Error
				do: 
					[:ex | 
					self logErrorBuildInputErrorFrom: eachRow reason: ex.
					inputWalkbacksWritten &gt; 2 
						ifFalse: 
							[PvWalkbackWriter new dumpWalkbackForException: ex.
							inputWalkbacksWritten := inputWalkbacksWritten + 1]]].
	self system logDebugTwo: #PLAN_STAT text: 'Loaded ' , inputCount asString , ' inputs from disk'.

	"Actually assign each ts it's label"
	self rawData do: 
		[:ts | 
		ts streamLabel assignToTimeSeries: ts]</body>
</methods>
<!-- Package Changes for PV-CME-Core(Kila.118,apikovsky)* -->


<methods>
<class-id>PROVISO.CmeESTreeBuilder</class-id> <category>building buckets</category>

<body package="PV-CME-Core" selector="buildForRequest:id:explicit:">buildForRequest: aRequest id: aTSid explicit: aBoolean 

	| ts tsgen request explicit returnSpecific specificId genericId |
	(specificId := self getActualIdForRequest: aRequest id: aTSid) ifNil: [^nil].
	ts := nil.
	tsgen := nil.
	returnSpecific := true.
	request := aRequest.
	explicit := aBoolean.
	(specificId = aTSid and: [specificId formula isGenericFormula])
		ifTrue: [self specificIdAndRequestForGenericTimeSeriesId: specificId
				do: 
					[:myId :myReq |
					genericId := specificId.
					specificId := myId.
					request := myReq.
					explicit := true.
					returnSpecific := false]].
	
	[ts := self
				buildOperatorForRequest: request
				id: specificId
				explicit: explicit.
	tsgen := specificId formula isSpecificToGenericConversion
				ifTrue: [self
						buildForGenericFormulaForRequest: request
						timeSeries: ts
						explicit: explicit]
				ifFalse: [ts].
	specificId formula isLocalizedSchedule
		ifTrue: [(self getHeartbeatSourceAt: specificId formula name)
				subscribe: ts]
		ifFalse: [aRequest currentResource notNil
				ifTrue: [self buildRaBucketsForTimeSeries: tsgen contextResource: request currentResource]
				ifFalse: [nil]]]
		on: Error
		do: 
			[:ex | 
			self
				logBuildErrorForRequest: request
				id: aTSid
				reason: ex.
			(self system debugFlags 
				at: #DEBUG_WALKBACK_ON_BUILD_ERROR
				ifAbsent: [true]) 
					ifTrue: [PvWalkbackWriter new dumpWalkbackForException: ex]].
	^returnSpecific
		ifTrue: [ts ifNil: [self tupleRepository at: specificId]]
		ifFalse:[tsgen ifNil: [self tupleRepository at: genericId]]</body>

<body package="PV-CME-Core" selector="specificIdAndRequestForGenericTimeSeriesId:do:">specificIdAndRequestForGenericTimeSeriesId: genId do: aBlock

	| specificRequests specificRequest specificId |
	genId formula isGenericFormula ifFalse: [^nil].
	specificRequests := nil.
	genId resource metricRequestsDo: 
		[:req | 
			(genId formula specificFormulas anySatisfy: [:eachFormula | eachFormula isSnmpFormula not and: [eachFormula = req formula]])
				ifTrue: 
					[specificRequests ifNil: [specificRequests := IdentitySet new].
					specificRequests add: req]].
	specificRequests ifNil: 
		["self logNoRequestsForSpecificFormulaForGenericTimeSeries: genId."
		^nil].

	specificRequests size &gt; 1 
		ifTrue: [self logTooManySpecificFormulaRequests: specificRequests forGenericTimeSeries: genId].

	(specificRequest := specificRequests anyOne) currentResource: genId resource.
	specificId := MetricSeriesIdentifier metric: specificRequest formula resource: genId resource.
	aBlock value: specificId value: specificRequest</body>
</methods>

<methods>
<class-id>PROVISO.CmeESTreeBuilder</class-id> <category>logging</category>

<body package="PV-CME-Core" selector="logNoRequestsForSpecificFormulaForGenericTimeSeries:">logNoRequestsForSpecificFormulaForGenericTimeSeries: aTimeSeries 
	
	self system
		logWarning: #NO_SPECIFIC_FORMULA_REQUEST
		messageId: #DC10011
		text: 'There is no specific formula request for timeseries: ' , aTimeSeries printString</body>

<body package="PV-CME-Core" selector="logTooManySpecificFormulaRequests:forGenericTimeSeries:">logTooManySpecificFormulaRequests: aCollOfReqs forGenericTimeSeries: aTimeSeries

	self system
		logWarning: #TOO_MANY_SPECIFIC_FORMULA_REQUESTS
		messageId: #DC10012
		text: 'There are too many specific formula requests for ' , aTimeSeries printString , ' ' , aCollOfReqs asArray printString , '.  This is a misconfiguration that can cause loss of data and excessive plan churn'.</body>
</methods>

<class>
<name>MockRequestEnumerator</name>
<environment>PROVISO</environment>
<super>PROVISO.CmeESTreeBuilder</super>
<private>false</private>
<indexed-type>none</indexed-type>
<inst-vars>requests matchBlock prunerClass </inst-vars>
<class-inst-vars></class-inst-vars>
<imports></imports>
<category>PV-CME-ES</category>
<attributes>
<package>PV-CME-Core</package>
</attributes>
</class>

<methods>
<class-id>PROVISO.MockRequestEnumerator</class-id> <category>initialize-release</category>

<body package="PV-CME-Core" selector="initialize">initialize
	
	super initialize.
	requests := OrderedCollection new.
	self beForAssociations.
	prunerClass := CmeESTreePruner</body>
</methods>

<methods>
<class-id>PROVISO.MockRequestEnumerator</class-id> <category>enumerating</category>

<body package="PV-CME-Core" selector="do:">do: aBlock 

	self requests do: 
		[:request | 
		aBlock 
			value: request
			value: request target]</body>
</methods>

<methods>
<class-id>PROVISO.MockRequestEnumerator</class-id> <category>building</category>

<body package="PV-CME-Core" selector="build">build

	PvFormulaLinker linkFormulas: (self requests collect: [:e | e formula]) inPvModel: self pvModel.
	super build</body>
</methods>

<methods>
<class-id>PROVISO.MockRequestEnumerator</class-id> <category>private</category>

<body package="PV-CME-Core" selector="newRequestForFormula:resource:">newRequestForFormula: aFormula resource: aSubelement

	| request |
	request := 
		PvRequest new
			id: self pvModel getNextId; 
			targetType: 'alias'; 
			target: aSubelement; 
			targetIndex: aSubelement id; 
			period: 900; 
			status: 'active'; 
			storageState: 'always'; 
			origin: 'Smalltalk'; 
			formulaIndex: aFormula id; 
			formula: aFormula; 
			yourself.
	self pvModel requestManager addItem: request.
	^request</body>
</methods>

<methods>
<class-id>PROVISO.MockRequestEnumerator</class-id> <category>accessing</category>

<body package="PV-CME-Core" selector="associations:">associations: aCollection 
	
	self pvModel requestManager
		initialize; 
		relink.
	self requests: (aCollection collect: [:assoc | self newRequestForFormula: assoc key resource: assoc value])</body>

<body package="PV-CME-Core" selector="beForAssociations">beForAssociations
	
	matchBlock := 
		[:arr :node | 
		self requests noneSatisfy: 
			[:eachRequest | eachRequest formula id = arr first and: [eachRequest target id = arr last]]]</body>

<body package="PV-CME-Core" selector="beForIncrementalTreeBuild">beForIncrementalTreeBuild

	self matchBlock: [:arr :bucket | 
		self pvModel resyncAgent dirtySubelements anySatisfy: [:eachSe | eachSe id = arr last]].</body>

<body package="PV-CME-Core" selector="emulateCmeTreeBulder">emulateCmeTreeBulder
	
	self matchBlock: 
		[:arr :bucket | 
		self fullTreeBuildFlag or: 
			[self pvModel resyncAgent dirtySubelements anySatisfy: 
				[:eachSe | eachSe id = arr last]]]</body>

<body package="PV-CME-Core" selector="formulas">formulas 
	
	^IdentitySet new addAll: (self requests collect: [:each | each formula])</body>

<body package="PV-CME-Core" selector="replaceAssociationsWith:">replaceAssociationsWith: aCollection

	self associations: aCollection</body>

<body package="PV-CME-Core" selector="requests">requests

	^requests</body>

<body package="PV-CME-Core" selector="requests:">requests: aCollection

	requests := aCollection</body>

<body package="PV-CME-Core" selector="resources">resources

	^IdentitySet new
		addAll: (self requests collect: [:each | each target]); 
		yourself</body>
</methods>

<remove-selector>
<class-id>PROVISO.MockRequestEnumerator</class-id> <selector>timeSeriesToRequests:</selector>
</remove-selector>

<remove-selector>
<class-id>PROVISO.MockRequestEnumerator</class-id> <selector>timeSeriesToRequests</selector>
</remove-selector>

<remove-selector>
<class-id>PROVISO.MockRequestEnumerator</class-id> <selector>associationsToIgnore:</selector>
</remove-selector>

<remove-selector>
<class-id>PROVISO.MockRequestEnumerator</class-id> <selector>newRequestForFormula:</selector>
</remove-selector>

<remove-selector>
<class-id>PROVISO.MockRequestEnumerator</class-id> <selector>associations</selector>
</remove-selector>

<remove-selector>
<class-id>PROVISO.MockRequestEnumerator</class-id> <selector>associationsToIgnore</selector>
</remove-selector>

<remove-selector>
<class-id>PROVISO.MockRequestEnumerator</class-id> <selector>associateTimeSeries:withRequest:</selector>
</remove-selector>

<reorganize>
<class-id>PROVISO.MockRequestEnumerator</class-id> <organization>('initialize-release' #initialize)
('enumerating' #do:)
('testing' #matchesTimeSeries:)
('building' #build)
('private' #considerNodesToUpdate #newRequestForFormula:resource:)
('accessing' #associations: #beForAssociations #beForIncrementalTreeBuild #emulateCmeTreeBulder #formulas #matchAll #matchBlock #matchBlock: #prunerClass #prunerClass: #replaceAssociationsWith: #requests #requests: #resources)
</organization>
</reorganize>
<!-- Package Changes for PV-ES-Formulas(Kila.72,apikovsky)* -->


<methods>
<class-id>PROVISO.LocalizedSchedule</class-id> <category>testing</category>

<body package="PV-ES-Formulas" selector="isGenericFormula">isGenericFormula

	^false</body>
</methods>
<!-- Package Changes for PV-DM-Support(Kila.73,apikovsky)* -->


<methods>
<class-id>PROVISO.PvFormula</class-id> <category>accessing</category>

<body package="PV-DM-Support" selector="addSpecificFormula:">addSpecificFormula: aFormula

	self specificFormulas ifNil: [self specificFormulas: Set new].
	self specificFormulas add: aFormula</body>

<body package="PV-DM-Support" selector="resetSpecificFormulas">resetSpecificFormulas

	self specificFormulas: nil</body>
</methods>

<methods>
<class-id>PROVISO.PvFormulaVisitor</class-id> <category>visitor-callbacks</category>

<body package="PV-DM-Support" selector="acceptFormula:">acceptFormula: aPvFormula 
	
	self pushFormula: aPvFormula.
	[(aPvFormula isOperatorFormulaOrProperty and: [aPvFormula formulaClass isNil]) 
		ifTrue: 
			[PvFormulaCompileNotification 
				signalWith: aPvFormula
				text: 'Missing formula prototype -- probably syntax error'].
	aPvFormula isSpecificToGenericConversion 
		ifTrue: 
			[(self pvModel formulaWithId: aPvFormula genericId) 
				ifNotNil: [:genericFormula | genericFormula addSpecificFormula: aPvFormula]].
	super acceptFormula: aPvFormula] 
		ensure: [self popFormula]</body>
</methods>

<methods>
<class-id>PROVISO.PvFormulaLinker</class-id> <category>api</category>

<body package="PV-DM-Support" selector="link">link
	self class allSmeeFormulasIn: self pvModel do: [:each | each resetFormulaConnections].
	self pvModel formulaManager itemsDo: [:formula | formula resetSpecificFormulas].
	self visit: self pvModel formulaManager</body>
</methods>
<!-- Package Changes for PV-Discovery-Core(Kila.1,hbolla)* -->


<methods>
<class-id>PROVISO.PvDiscoveryPipeline</class-id> <category>logging</category>

<body package="PV-Discovery-Core" selector="logProfileStucked">logProfileStucked
	
	self system 
		logWarning: #PROFILENOTPROGRESSING
		messageId: #DL39001
		text: self tag , '. Profile is not progressing during the last ' , self stuckTime asString, ' seconds, aborting ...'.</body>
</methods>
<!-- Package Changes for PV-CFL-UI(Kila.31,apikovsky)* -->


<methods>
<class-id>PROVISO.PvESFormulaEditor</class-id> <category>actions</category>

<body package="PV-CFL-UI" selector="buildComputation">buildComputation
	
	| treeBuilder |
	self validateBeforeBuild 
		ifFalse: 
			[self statusFailure: 'Formula not deployed'.
			^self].
	self results: OrderedCollection new.
	self removeBuildErrors.
	self resetTupleRepository.
	self environment do: 
		[treeBuilder := MockRequestEnumerator new.
		treeBuilder
			environment: Smee.SmeeEnvironment current;
			tupleRepository: self tupleRepository;
			shouldBuildRaAndGaBuckets: false;
			pvModel: self pvModel;
			collectorNumber: 1.
		treeBuilder associations: (Array with: self selectedFormula -&gt; self selectedSubelement).
		treeBuilder addDependent: self.
		[[treeBuilder build] 
			on: Error
			do: [:ex | ^Dialog notify: 'Fatal Build Error: ' , ex description]] 
				on: TreeBuildError
				do: 
					[:ex1 | 
					self addBuildError: ex1.
					ex1 resume]].
	self updateInputChoices.
	self updateStateOfTimeSeriesDependentControls.
	self statusSuccess: 
		'Ready for testing, click ''Show Tree'' at any time to see details'</body>
</methods>
<!-- Package Changes for PV-Base Extensions(Kila.30,jhylands)* -->


<methods>
<class-id>Core.Symbol</class-id> <category>accessing</category>

<body package="PV-Base Extensions" selector="findNextUnusedMessageIdForDC">findNextUnusedMessageIdForDC
	"DC messages start at 10000"
	^#DC10000 privateFindNextUnusedMessageId</body>

<body package="PV-Base Extensions" selector="findNextUnusedMessageIdForDL">findNextUnusedMessageIdForDL
	"DL messages start at 39000"

	^#DL39000 privateFindNextUnusedMessageId</body>

<body package="PV-Base Extensions" selector="privateFindNextUnusedMessageId">privateFindNextUnusedMessageId
	"
	#DC10000 privateFindNextUnusedMessageId
	#DL39000 privateFindNextUnusedMessageId
	"

	| firstNum letters number |
	firstNum := (1 to: self size) detect: [ :i | (self at: i) isDigit] ifNone: [(self size + 1)].
	letters := self copyFrom: 1 to: (firstNum - 1).
	number := (self copyFrom: firstNum to: self size) asNumber.

	[(Symbol findInterned: (letters, number asString)) ifNil: [^(letters, number asString)].
	number := number + 1] repeat.</body>
</methods>

<remove-selector>
<class-id>Core.Symbol</class-id> <selector>findNextUnusedMessageId</selector>
</remove-selector>


</st-source>
